import argparse
import logging
import os
from datetime import datetime
from typing import Optional

from .api import DAPClient
from .arguments import EnumAction, environment_default, valid_utc_datetime
from .dap_types import Format, IncrementalQuery, SnapshotQuery


class Arguments(argparse.Namespace):
    base_url: str
    client_id: str
    client_secret: str
    loglevel: str

    namespace: str
    table: str
    format: Format
    output_directory: str

    since: datetime
    until: Optional[datetime]


def log_level_name(level: int) -> str:
    name: str = logging.getLevelName(level)
    return name.lower()


# subcommand functions
def add_query_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--namespace",
        default="canvas",
        help="Identifies the data source.",
    )
    parser.add_argument(
        "--table",
        required=True,
        help="Table name whose data to fetch.",
    )
    parser.add_argument(
        "--format",
        type=Format,
        action=EnumAction,
        default=Format.JSONL,
        help="Data output format.",
    )
    parser.add_argument(
        "--output-directory",
        metavar="DIR",
        default="downloads",
        help="Directory where the query result will be downloaded to. Can be an absolute or relative path.",
    )


def parse_snapshot(args: Arguments) -> SnapshotQuery:
    return SnapshotQuery(format=args.format, filter=None)


def parse_incremental(args: Arguments) -> IncrementalQuery:
    return IncrementalQuery(
        format=args.format, filter=None, since=args.since, until=args.until
    )


def console_entry():
    parser = argparse.ArgumentParser(
        description="Invokes the DAP API to fetch table snapshots and incremental updates.",
        epilog="For more information, check out the OpenAPI specification for DAP API: https://data-access-platform-api.s3.amazonaws.com/index.html",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.prog = os.path.basename(os.path.dirname(__file__))

    parser.add_argument(
        "--base-url",
        metavar="URL",
        help="Base URL of the DAP API.",
        action=environment_default("DAP_API_URL"),
    )
    parser.add_argument(
        "--client-id",
        metavar="ID",
        help="Client ID.",
        action=environment_default("DAP_CLIENT_ID"),
    )
    parser.add_argument(
        "--client-secret",
        metavar="SECRET",
        help="Client secret.",
        action=environment_default("DAP_CLIENT_SECRET"),
    )
    parser.add_argument(
        "--loglevel",
        choices=[
            log_level_name(level)
            for level in (
                logging.DEBUG,
                logging.INFO,
                logging.WARN,
                logging.ERROR,
                logging.CRITICAL,
            )
        ],
        default=log_level_name(logging.INFO),
        help="Sets log verbosity.",
    )

    subparsers = parser.add_subparsers(
        help="Command to execute, e.g. initiate a snapshot or and incremental query.",
        required=True,
    )

    snapshot_parser = subparsers.add_parser(
        "snapshot", help="Performs a snapshot query."
    )
    add_query_arguments(snapshot_parser)
    snapshot_parser.set_defaults(parse_query=parse_snapshot)

    incremental_parser = subparsers.add_parser(
        "incremental",
        help="Performs an incremental query with a given start, and (optionally) end timestamp.",
    )
    add_query_arguments(incremental_parser)
    incremental_parser.add_argument(
        "--since",
        metavar="DATETIME",
        required=True,
        help="Start timestamp for an incremental query. Examples: 2022-06-13T09:30:00Z or 2022-06-13T09:30:00+02:00.",
        type=valid_utc_datetime,
    )
    incremental_parser.add_argument(
        "--until",
        metavar="DATETIME",
        required=False,
        help="End timestamp for an incremental query. Examples: 2022-06-13T09:30:00Z or 2022-06-13T09:30:00+02:00.",
        type=valid_utc_datetime,
    )
    incremental_parser.set_defaults(parse_query=parse_incremental)

    args = Arguments()
    parser.parse_args(namespace=args)

    logging.basicConfig(
        level=getattr(logging, args.loglevel.upper(), logging.INFO),
        format="%(asctime)s - %(levelname)s - %(funcName)s [%(lineno)d] - %(message)s",
    )

    query = args.parse_query(args)

    with DAPClient(
        base_url=args.base_url,
        client_id=args.client_id,
        client_secret=args.client_secret,
    ) as my_client:
        my_client.fetch_table_data(
            args.namespace, args.table, query, args.output_directory
        )


if __name__ == "__main__":
    console_entry()
