import logging
import os
import shutil
import time
from typing import Any, List, Type, TypeVar
from urllib.parse import urlparse

import requests
from strong_typing.serialization import json_dump_string, json_to_object, object_to_json

from .dap_types import (
    Credentials,
    IncrementalQuery,
    Job,
    JobID,
    JobStatus,
    Object,
    Query,
    Resource,
    ResourceResult,
    SnapshotQuery,
    TableJob,
    TokenProperties,
)

T = TypeVar("T")

credential_type = "Canvas"


class DAPClientError(RuntimeError):
    pass


class DAPClient:
    _base_url: str
    _client_id: str
    _client_secret: str
    _session: "DAPSession"

    def __init__(
        self, base_url: str = None, client_id: str = None, client_secret: str = None
    ) -> None:
        base_url_arg = base_url or os.getenv("DAP_API_URL")
        client_id_arg = client_id or os.getenv("DAP_CLIENT_ID")
        client_secret_arg = client_secret or os.getenv("DAP_CLIENT_SECRET")

        if not base_url_arg:
            raise DAPClientError("base URL missing")
        if not client_id_arg:
            raise DAPClientError("client ID missing")
        if not client_secret_arg:
            raise DAPClientError("client secret missing")

        self._base_url = base_url_arg
        self._client_id = client_id_arg
        self._client_secret = client_secret_arg

        logging.info(f"Client ID: {self._client_id}")

    def __enter__(self):
        session = requests.Session()
        credentials = Credentials(credential_type, self._client_id, self._client_secret)
        self._session = DAPSession(session, self._base_url)
        try:
            logging.info(f"Connecting to {self._base_url}")
            self._session.authenticate(credentials)
        except Exception as e:
            logging.error(f"Failed to authenticate: {e}")
            self._session.close()
            self._session = None
            raise
        return self._session

    def __exit__(self, type, value, traceback):
        self._session.close()
        self._session = None


class DAPSession:
    "Represents an authenticated session to DAP."

    _session: requests.Session

    def __init__(self, session: requests.Session, base_url: str) -> None:
        "Creates a new logical session by encapsulating a network connection."

        self._base_url = base_url
        self._session = session

    def close(self) -> None:
        "Closes the underlying network socket."

        self._session.close()

    def _get(self, path: str, response_type: Type[T]) -> T:
        "Sends a request to the server and parses the response into the expected type."

        response = self._session.get(f"{self._base_url}{path}")
        response.raise_for_status()
        response_payload = response.json()
        logging.debug(f"GET response payload:\n{repr(response_payload)}")

        return json_to_object(response_type, response_payload)

    def _post(self, path: str, request_data: Any, response_type: Type[T]) -> T:
        "Sends a request to the server by serializing a payload object, and parses the response into the expected type."

        request_payload = object_to_json(request_data)
        logging.debug(f"POST request payload:\n{repr(request_payload)}")

        response = self._session.post(
            f"{self._base_url}{path}",
            data=json_dump_string(request_payload),
            headers={"Content-Type": "application/json"},
        )
        response.raise_for_status()
        response_payload = response.json()
        logging.debug(f"POST response payload:\n{repr(response_payload)}")

        return json_to_object(response_type, response_payload)

    def authenticate(self, credentials: Credentials) -> None:
        "Authenticates to DAP with client ID and secret to receive a JWT."

        logging.info(f"Authenticating to DAP with {credentials.client_id}")
        properties = self._post("/auth", credentials, TokenProperties)
        self._session.headers.update(
            {"Authorization": f"Bearer {properties.access_token}"}
        )

    def query_snapshot(self, namespace: str, table: str, query: SnapshotQuery) -> Job:
        "Start a snapshot query."

        logging.info(f"Query snapshot of table: {table}")
        job = self._post(f"/query/{namespace}/table/{table}/data", query, TableJob)
        return job

    def query_incremental(
        self, namespace: str, table: str, query: IncrementalQuery
    ) -> Job:
        "Start an incremental query."

        logging.info(f"Query updates for table: {table}")
        job = self._post(f"/query/{namespace}/table/{table}/data", query, TableJob)
        return job

    def get_job(self, job_id: JobID) -> Job:
        "Retrieve job status."

        logging.debug(f"Retrieving job state for job {job_id}")
        job = self._get(f"/job/{job_id}", Job)  # type: ignore
        return job

    def get_job_status(self, job_id: JobID) -> JobStatus:
        "Retrieve job status."

        return self.get_job(job_id).status

    def get_objects(self, job_id: JobID) -> List[Object]:
        "Retrieve object IDs once the query is completed successfully."

        logging.info(f"Retrieving object IDs for job {job_id}")
        job = self._get(f"/job/{job_id}", Job)  # type: ignore
        return job.objects

    def get_resources(self, objects: List[Object]) -> List[Resource]:
        "Retrieve URLs to data stored remotely."

        object_id_list = ", ".join(o.id for o in objects)
        logging.info(f"Retrieve resource URLs for objects: {object_id_list}")
        response = self._post("/object/url", objects, ResourceResult)
        resource_list = [response.urls[resource_id] for resource_id in response.urls]
        return resource_list

    def download_resources(
        self, resources: List[Resource], job_id: JobID, output_directory: str
    ) -> List[str]:
        """
        Save data stored remotely into a local directory.

        :param output_directory: Path to the directory to save downloaded files to.
        :returns: A list of paths to files saved in the local file system.
        """

        local_files: List[str] = []
        dir = os.path.join(output_directory, f"job_{job_id}")
        os.makedirs(dir, exist_ok=True)
        for u in resources:
            url = str(u.url)
            url_path = urlparse(url).path
            file_base_name = os.path.basename(url_path)
            file_path = os.path.join(dir, file_base_name)
            with requests.get(url, stream=True) as data, open(file_path, "wb") as file:
                # save gzip data to file without decompressing
                shutil.copyfileobj(data.raw, file)
            local_files.append(file_path)
        logging.info(f"Files from server downloaded to folder: {dir}")
        return local_files

    def await_job(self, job: Job) -> Job:
        """
        Wait until a job terminates.

        :param job: A job that might be still running.
        :returns: A job that has completed with success or terminated with failure.
        """

        while job.status is JobStatus.Running:
            delay = 30
            logging.info(
                f"Query job still running, checking again in {delay} seconds..."
            )
            time.sleep(delay)

            job = self.get_job(job.id)

        logging.info(f"Query job finished with status: {job.status.value}")
        return job

    def execute_job(
        self,
        namespace: str,
        table: str,
        query: Query,
    ) -> Job:
        "Start a query job and wait until it terminates."

        if isinstance(query, SnapshotQuery):
            job = self.query_snapshot(namespace, table, query)
        elif isinstance(query, IncrementalQuery):
            job = self.query_incremental(namespace, table, query)
        else:
            raise TypeError(f"type mismatch for parameter `query`: {type(query)}")

        logging.info(f"Query started with job ID: {job.id}")

        job = self.await_job(job)
        return job

    def fetch_table_data(
        self, namespace: str, table: str, query: Query, output_directory: str
    ) -> List[str]:
        """
        Execute a query job and download data to a local directory.

        :param output_directory: Path to the directory to save downloaded files to.
        :returns: A list of paths to files saved in the local file system.
        """

        # fail early if output directory does not exist and cannot be created
        os.makedirs(output_directory, exist_ok=True)

        job = self.execute_job(namespace, table, query)

        if job.status is JobStatus.Complete:
            objects = self.get_objects(job.id)
            resources = self.get_resources(objects)
        else:
            raise DAPClientError(f"Query job ended with status: {job.status.value}")

        return self.download_resources(resources, job.id, output_directory)
